from typing import Dict, Set

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class Offset(PrivateSingleChildOpenScadCommand):
    """
    Offset generates a new 2D interior or exterior outline from an existing outline. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#offset.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 radius: float | None = None,
                 delta: float | None = None,
                 chamfer: bool | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param radius: The radius of the circle that is rotated about the outline, either inside or outside. This mode
                       produces rounded corners.
        :param delta: The distance of the new outline from the original outline, and therefore reproduces angled
                      corners. No inward perimeter is generated in places where the perimeter would cross itself.
        :param chamfer: When using the delta parameter, this flag defines if edges should be chamfered (cut off with a
                        straight line) or not (extended to their intersection). This parameter has no effect on radial
                        offsets.
        :param child: The child widget.
        """
        if delta is not None:
            if chamfer is None:
                chamfer = False
        else:
            chamfer = None

        PrivateSingleChildOpenScadCommand.__init__(self, command='offset', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'radius': 'r', 'fa': '$fa', 'fs': '$fs', 'fn': '$fn'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'r', 'delta', '$fs'}

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def radius(self) -> float | None:
        """
        Returns the radius.
        """
        return self._args.get('radius')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def delta(self) -> float | None:
        """
        Returns the delta.
        """
        return self._args.get('delta')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def chamfer(self) -> bool | None:
        """
        Returns the delta.
        """
        return self._args.get('chamfer')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._args.get('fa')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._args.get('fs')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._args.get('fn')

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.private.PrivateMultiChildOpenScadCommand import PrivateMultiChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class Difference(PrivateMultiChildOpenScadCommand):
    """
    Subtracts the second (and all further) child widgets from the first child widgets (logical and not).
    See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/CSG_Modelling#difference.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, children: List[ScadWidget]):
        """
        Object constructor.

        :param children: The child widgets.
        """
        PrivateMultiChildOpenScadCommand.__init__(self, command='difference', args=locals(), children=children)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateScale import PrivateScale
from super_scad.type.Vector2 import Vector2


class Scale2D(ScadSingleChildParent):
    """
    Scales its child widget using a specified scaling factor.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 factor: Vector2 | float | None = None,
                 factor_x: float | None = None,
                 factor_y: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param factor: The scaling factor along all the two axes.
        :param factor_x: The scaling factor along the x-axis.
        :param factor_y: The scaling factor along the y-axis.
        :param child: The child to be scaled.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._factor: Vector2 | float | None = factor
        """
        The scaling factor along all the two axes.
        """

        self._factor_x: float | None = factor_x
        """
        The scaling factor along the x-axis.
        """

        self._factor_y: float | None = factor_y
        """
        The scaling factor along the y-axis.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'factor'}, {'factor_x', 'factor_y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def factor(self) -> Vector2:
        """
        Returns the scaling factor along all two axes.
        """
        if self._factor is None:
            self._factor = Vector2(self.factor_x, self.factor_y)
        elif isinstance(self._factor, float):
            self._factor = Vector2(self._factor, self._factor)

        return self._factor

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def factor_x(self) -> float:
        """
        Returns the scaling factor along the x-axis.
        """
        if self._factor_x is None:
            if isinstance(self._factor, Vector2):
                self._factor_x = self._factor.x
            elif isinstance(self._factor, float):
                self._factor_x = self._factor
            else:
                self._factor_x = 1.0

        return self._factor_x

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def factor_y(self) -> float:
        """
        Returns the scaling factor along the y-axis.
        """
        if self._factor_y is None:
            if isinstance(self._factor, Vector2):
                self._factor_y = self._factor.y
            elif isinstance(self._factor, float):
                self._factor_y = self._factor
            else:
                self._factor_y = 1.0

        return self._factor_y

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateScale(factor=self.factor, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, List, Set

from super_scad.d2.PolygonMixin import PolygonMixin
from super_scad.d2.private.PrivateSquare import PrivateSquare
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2


class Rectangle(PolygonMixin, ScadWidget):
    """
    Widget for creating rectangles.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: Vector2 | None = None,
                 width: float | None = None,
                 depth: float | None = None,
                 center: bool = False,
                 extend_by_eps_sides: bool | List[bool] | Set[int] | None = None):
        """
        Object constructor.

        :param size: The size of the rectangle.
        :param width: The width (the size along the x-axis) of the rectangle.
        :param depth: The depth (the size along the y-axis) of the rectangle.
        :param center: Whether the rectangle is centered at its position.
        :param extend_by_eps_sides: Whether to extend sides by eps for a clear overlap.
        """
        ScadWidget.__init__(self)
        PolygonMixin.__init__(self, extend_by_eps_sides=extend_by_eps_sides)

        self._size: Vector2 | None = size
        """
        The size of the rectangle.
        """

        self._width: float | None = width
        """
        The width of the rectangle.
        """

        self._depth: float | None = depth
        """
        The depth of the rectangle.
        """

        self._center: bool = center
        """
        Whether the rectangle center is at the center of the rectangle.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'size'}, {'width', 'depth'})
        validator.validate_required({'width', 'size'},
                                    {'depth', 'size'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> Vector2:
        """
        Returns the size of this rectangle.
        """
        if self._size is None:
            self._size = Vector2(self._width, self._depth)

        return self._size

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def width(self) -> float:
        """
        Returns the width (the size along the x-axis) of this rectangle.
        """
        if self._width is None:
            self._width = self._size.x

        return self._width

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def depth(self) -> float:
        """
        Returns the depth (the size along the y-axis) of this rectangle.
        """
        if self._depth is None:
            self._depth = self._size.y

        return self._depth

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the rectangle is centered at this origin.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def sides(self) -> int:
        """
        Returns the number of sides of this rectangle.
        """
        return 4

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def nodes(self) -> List[Vector2]:
        """
        Returns the nodes of this rectangle.
        """
        if self.center:
            return [Vector2(-0.5 * self.width, -0.5 * self.depth),
                    Vector2(-0.5 * self.width, 0.5 * self.depth),
                    Vector2(0.5 * self.width, 0.5 * self.depth),
                    Vector2(0.5 * self.width, -0.5 * self.depth)]

        return [Vector2.origin,
                Vector2(0.0, self.depth),
                Vector2(self.width, self.depth),
                Vector2(self.width, 0.0)]

    # ------------------------------------------------------------------------------------------------------------------
    def _build_polygon(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateSquare(size=self.size, center=self.center)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivatePaint import PrivatePaint
from super_scad.type.Color import Color


class Paint(ScadSingleChildParent):
    """
    Paints a child widget using a specified color and opacity. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#color.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 color: Color | str | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param color: The color and opacity of the child widget.
        :param child: The child widget to be painted.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._color: Color | str | None = color
        """
        The color and opacity of the child widget.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'color'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def color(self) -> Color:
        """
        Returns the color and opacity of the child widget.
        """
        if isinstance(self._color, str):
            self._color = Color(self._color)

        return self._color

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivatePaint(color=self.color, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from abc import ABC
from typing import List

from super_scad.scad.ScadWidget import ScadWidget


class ScadMultiChildParent(ScadWidget, ABC):
    """
    Abstract widget for creating SuperSCAD widgets that have multiple children.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, children: List[ScadWidget]):
        """
        Object constructor.

        :param children: The child SuperSCAD widgets of this multi-child parent.
        """
        ScadWidget.__init__(self)

        self.__children: List[ScadWidget] = children
        """
        The child OpenSCAD widgets of this multi-child parent.
        """

        for key, child in enumerate(self.children):
            assert isinstance(child, ScadWidget), f"Child {key} is of type: {child.__class__}"

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def children(self) -> List[ScadWidget]:
        """
        Returns the children of this multi-child parent.
        """
        return self.__children

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, List, Set

from super_scad.d2.PolygonMixin import PolygonMixin
from super_scad.d2.private.PrivateSquare import PrivateSquare
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type import Vector2


class Square(PolygonMixin, ScadWidget):
    """
    Widget for creating squares.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: float,
                 center: bool = False,
                 extend_by_eps_sides: bool | List[bool] | Set[int] | None = None):
        """
        Object constructor.

        :param size: The size of the square.
        :param center: Whether the square is centered at the origin.
        :param extend_by_eps_sides: Whether to extend sides by eps for a clear overlap.
        """
        ScadWidget.__init__(self)
        PolygonMixin.__init__(self, extend_by_eps_sides=extend_by_eps_sides)

        self._size: float = size
        """
        The size of the square.
        """

        self._center: bool = center
        """
        Whether the square is centered at the origin.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_required({'size'}, {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> float:
        """
        Returns the size of this square.
        """
        return self._size

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether this square is centered at its position.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def sides(self) -> int:
        """
        Returns the number of sides of this square.
        """
        return 4

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def nodes(self) -> List[Vector2]:
        """
        Returns the nodes of this square.
        """
        if self.center:
            return [Vector2(-0.5 * self.size, -0.5 * self.size),
                    Vector2(-0.5 * self.size, 0.5 * self.size),
                    Vector2(0.5 * self.size, 0.5 * self.size),
                    Vector2(0.5 * self.size, -0.5 * self.size)]

        return [Vector2.origin,
                Vector2(0.0, self.size),
                Vector2(self.size, self.size),
                Vector2(self.size, 0.0)]

    # ------------------------------------------------------------------------------------------------------------------
    def _build_polygon(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateSquare(size=self.size, center=self.center)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateRotate import PrivateRotate
from super_scad.type.Vector2 import Vector2


class Flip2D(ScadSingleChildParent):
    """
    Flips its child about the x or y-axis.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 horizontal: bool | None = None,
                 vertical: bool | None = None,
                 flip_x: bool | None = None,
                 flip_y: bool | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param horizontal: Whether to flip the child widget horizontally (i.e., flip around the y-axis).
        :param vertical: Whether to flip the child widget vertically (i.e., flip around the x-axis).
        :param flip_x: Whether to flip the child widget around the x-asis (i.e., vertical flip).
        :param flip_y: Whether to flip the child widget around the y-asis (i.e., horizontal flip).
        :param child: The child widget to be flipped.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._horizontal: bool | None = horizontal
        """
        Whether to flip the child widget horizontally (i.e., flip around the y-axis).
        """

        self._vertical: bool | None = vertical
        """
        Whether to flip the child widget vertically (i.e., flip around the x-axis).
        """

        self._flip_x: bool | None = flip_x
        """
        Whether to flip the child widget around the x-asis (i.e., vertical flip).
        """

        self._flip_y: bool | None = flip_y
        """
        Whether to flip the child widget around the y-asis (i.e., horizontal flip).
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'horizontal', 'vertical'}, {'flip_x', 'flip_y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def horizontal(self) -> bool:
        """
        Returns whether to flip the child widget horizontally (i.e., flip around the y-axis).
        """
        if self._horizontal is None:
            horizontal = self._horizontal or False
            flip_y = self._flip_y or False
            self._horizontal = horizontal or flip_y

        return self._horizontal

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vertical(self) -> bool:
        """
        Returns whether to flip the child widget vertically (i.e., flip around the x-axis).
        """
        if self._vertical is None:
            vertical = self._vertical or False
            flip_x = self._flip_x or False
            self._vertical = vertical or flip_x

        return self._vertical

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def flip_x(self) -> bool:
        """
        Returns whether to flip the child widget around the x-asis (i.e., vertical flip).
        """
        if self._flip_x is None:
            self._flip_x = self.vertical or False

        return self._flip_x

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def flip_y(self) -> bool:
        """
        Returns whether to flip the child widget around the y-asis (i.e., horizontal flip).
        """
        if self._flip_y is None:
            self._flip_y = self.horizontal or False

        return self._flip_y

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        angle = Vector2(x=180.0 if self.flip_x else 0.0, y=180.0 if self.flip_y else 0.0)

        return PrivateRotate(angle=angle, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class Render(PrivateSingleChildOpenScadCommand):
    """
    Forces the generation of a mesh even in preview mode. This is useful in certain situations, e.g., when the boolean
    operations become too slow to track. Render can also be used (typically in conjunction with convexity) to
    avoid/workaround preview artifacts. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#render.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 convexity: int | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='render', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the
        child object.
        """
        return self._args.get('convexity')

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Rotate2D import Rotate2D
from super_scad.transformation.Translate2D import Translate2D
from super_scad.type import Vector2
from super_scad.type.Angle import Angle


class Position2D(ScadSingleChildParent):
    """
    A convenience widget that first rotates its child about the z-axis and then translates its child widget along the
    specified vector.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float | None = None,
                 vector: Vector2 | None = None,
                 x: float | None = None,
                 y: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param angle: The angle of rotation (around the z-axis).
        :param vector: The vector over which the child widget is translated.
        :param x: The distance the child widget is translated to along the x-axis.
        :param y: The distance the child widget is translated to along the y-axis.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._angle: float | None = angle
        """
        The angle of rotation (around the z-axis).
        """

        self._vector: Vector2 | None = vector
        """
        The vector over which the child widget is translated.
        """

        self._x: float | None = x
        """
        The distance the child widget is translated to along the x-axis.
        """

        self._y: float | None = y
        """
        The distance the child widget is translated to along the y-axis.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'vector'}, {'x', 'y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle(self) -> float:
        """
        Returns the angle of rotation (around the z-axis).
        """
        if self._angle is None:
            self._angle = 0.0

        return Angle.normalize(self._angle)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector2:
        """
        Returns the vector over which the child widget is translated.
        """
        if self._vector is None:
            self._vector = Vector2(self.x, self.y)

        return self._vector

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def x(self) -> float:
        """
        Returns distance the child widget is translated to along the x-axis.
        """
        if self._x is None:
            if self._vector is not None:
                self._x = self.vector.x
            else:
                self._x = 0.0

        return self._x

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def y(self) -> float:
        """
        Returns distance the child widget is translated to along the y-axis.
        """
        if self._y is None:
            if self._vector is not None:
                self._y = self.vector.y
            else:
                self._y = 0.0

        return self._y

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        child = self.child

        if self.angle != 0.0:
            child = Rotate2D(angle=self.angle, child=child)

        if self.vector.is_not_origin:
            child = Translate2D(vector=self.vector, child=child)

        return child

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, Set

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.scad.ArgumentValidator import ArgumentValidator


class PrivateSphere(PrivateOpenScadCommand):
    """
    Class for spheres. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#sphere.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 radius: float | None = None,
                 diameter: float | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None):
        """
        Object constructor.

        :param radius: The radius of the sphere.
        :param diameter: The diameter of the sphere.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        """
        PrivateOpenScadCommand.__init__(self, command='sphere', args=locals())

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'radius'}, {'diameter'})
        validator.validate_required({'radius', 'diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'radius': 'r', 'diameter': 'd', 'fa': '$fa', 'fs': '$fs', 'fn': '$fn'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return {'$fa'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'r', 'd', '$fs'}

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, List

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.scad.ScadMultiChildParent import ScadMultiChildParent
from super_scad.scad.ScadWidget import ScadWidget


class PrivateMultiChildOpenScadCommand(PrivateOpenScadCommand, ScadMultiChildParent):
    """
    Parent widget for OpenSCAD commands with multiple children.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, command: str, args: Dict[str, Any], children: List[ScadWidget]):
        """
        Object constructor.

        :param command: The OpenSCAD command.
        :param args: The arguments of the command.
        :param children: The child SuperSCAD widgets of this multi-child parent.
        """
        PrivateOpenScadCommand.__init__(self, command=command, args=args)
        ScadMultiChildParent.__init__(self, children=children)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, Tuple

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateResize import PrivateResize
from super_scad.type.Vector3 import Vector3


class Resize3D(ScadSingleChildParent):
    """
    Modifies the size of the child widget to match the given width, depth, and height. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#resize.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 new_size: Vector3 | None = None,
                 new_width: float | None = None,
                 new_depth: float | None = None,
                 new_height: float | None = None,
                 auto: bool | Tuple[bool, bool, bool] | None = None,
                 auto_width: bool | None = None,
                 auto_depth: bool | None = None,
                 auto_height: bool | None = None,
                 convexity: int | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param new_size: The new_size along all three axes.
        :param new_width: The new width (the new size along the x-axis).
        :param new_depth: The new depth (the new size along the y-axis).
        :param new_height: The new height size (the new size along the z-axis).
        :param auto: Whether to auto-scale any 0-dimensions to match.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        :param child: The child widget to be resized.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._new_size: Vector3 | None = new_size
        """
        The new_size along all three axes.
        """

        self._new_width: float | None = new_width
        """
        The new width (the new size along the x-axis).
        """

        self._new_depth: float | None = new_depth
        """
        The new depth (the new size along the y-axis).
        """

        self._new_height: float | None = new_height
        """ 
        The new height size (the new size along the z-axis).
        """

        self._auto: bool | Tuple[bool, bool, bool] | None = auto
        """
        Whether to auto-scale any 0-dimensions to match.
        """

        self._auto_width: bool | None = auto_width
        """
        Whether to auto-scale any 0-dimensions to match.
        """

        self._auto_depth: bool | None = auto_depth
        """
        Whether to auto-scale any 0-dimensions to match.
        """

        self._auto_height: bool | None = auto_height
        """
        Whether to auto-scale any 0-dimensions to match.
        """

        self._convexity: int | None = convexity
        """
        Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the child 
        widget.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'new_size'}, {'new_width', 'new_depth', 'new_height'})
        validator.validate_exclusive({'auto'}, {'auto_width', 'auto_depth', 'auto_height'})
        validator.validate_required({'new_width', 'new_size', 'auto_width', 'auto'},
                                    {'new_depth', 'new_size', 'auto_depth', 'auto'},
                                    {'new_height', 'new_size', 'auto_height', 'auto'})

        # Handle resizing beyond resolution.

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_size(self) -> Vector3:
        """
        Returns the new_size along all three axes.
        """
        if self._new_size is None:
            self._new_size = Vector3(self.new_width, self.new_depth, self.new_height)

        return self._new_size

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_width(self) -> float:
        """
        Returns new width (the new size along the x-axis).
        """
        if self._new_width is None:
            if self._new_size is not None:
                self._new_width = self._new_size.x
            else:
                self._new_width = 0.0

        return self._new_width

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_depth(self) -> float:
        """
        Returns the new depth (the new size along the y-axis).
        """
        if self._new_depth is None:
            if self._new_size is not None:
                self._new_depth = self._new_size.y
            else:
                self._new_depth = 0.0

        return self._new_depth

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_height(self) -> float:
        """
        Returns the new height (the new size along the z-axis).
        """
        if self._new_height is None:
            if self._new_size is not None:
                self._new_height = self._new_size.z
            else:
                self._new_height = 0.0

        return self._new_height

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto(self) -> Tuple[bool, bool, bool]:
        """
        Returns whether to auto-scale the width, depth and height.
        """
        return self.auto_width, self.auto_depth, self.auto_height

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto_width(self) -> bool:
        """
        Returns whether to auto-scale the width (the size along the x-axis).
        """
        if round(self.new_width, 4) == 0.0:  # xxx Use rounding in target units.
            if self._auto is not None:
                if isinstance(self._auto, tuple):
                    return self._auto[0]

                return self._auto

            if self._auto_width is not None:
                return self._auto_width

        return False

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto_depth(self) -> bool:
        """
        Returns whether to auto-scale the depth (the size along the y-axis).
        """
        if round(self.new_depth, 4) == 0.0:  # xxx Use rounding in target units.
            if self._auto is not None:
                if isinstance(self._auto, tuple):
                    return self._auto[1]

                return self._auto

            if self._auto_depth is not None:
                return self._auto_depth

        return False

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto_height(self) -> bool:
        """
        Returns whether to auto-scale the height (the size along the z-axis).
        """
        if round(self.new_height, 4) == 0.0:  # xxx Use rounding in target units.
            if self._auto is not None:
                if isinstance(self._auto, tuple):
                    return self._auto[2]

                return self._auto

            if self._auto_height is not None:
                return self._auto_height

        return False

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the convexity.
        """
        return self._convexity

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateResize(new_size=self.new_size, auto=self.auto, convexity=self.convexity, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



import inspect
import os
from pathlib import Path

from super_scad.scad import Length
from super_scad.scad.BOM import BOM
from super_scad.scad.ScadCodeStore import ScadCodeStore
from super_scad.scad.Unit import Unit
from super_scad.type import Vector3


class Context:
    """
    The context for generating OpenSCAD from SuperSCAD.
    """

    # ------------------------------------------------------------------------------------------------------------------
    DEFAULT_FA: float = 12.0
    """
    OpenSCAD default value for $fa.
    """

    DEFAULT_FS: float = 2.0
    """
    OpenSCAD default value for $fs.
    """

    DEFAULT_FN: int = 0
    """
    OpenSCAD default value for $fn.
    """

    DEFAULT_VIEWPORT_ROTATION: Vector3 = Vector3(55.0, 0.0, 25)
    """
    The default viewport rotation.
    """

    DEFAULT_VIEWPORT_TRANSLATION: Vector3 = Vector3.origin
    """
    The default viewport translation.
    """

    DEFAULT_VIEWPORT_DISTANCE: float = 140.0
    """
    The default FOV (Field Of View).
    """

    DEFAULT_VIEWPORT_FIELD_OF_VIEW: float = 22.5
    """
    The default FOV (Field Of View).
    """

    __unit_length_current: Unit = Unit.FREE
    """
    The current unit of length.
    """

    __unit_length_final: Unit = Unit.FREE
    """
    The unit of length used in the generated OpenSCAD code.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 unit_length_final: Unit = Unit.MM,
                 fa: float = DEFAULT_FA,
                 fs: float = DEFAULT_FS,
                 fn: int = DEFAULT_FN,
                 vpr: Vector3 = DEFAULT_VIEWPORT_ROTATION,
                 vpt: Vector3 = DEFAULT_VIEWPORT_TRANSLATION,
                 vpd: float = DEFAULT_VIEWPORT_DISTANCE,
                 vpf: float = DEFAULT_VIEWPORT_FIELD_OF_VIEW,
                 eps: float = 1E-2,
                 delta: float = 1e-5,
                 angle_digits: int = 4,
                 length_digits: int = 4,
                 scale_digits: int = 4):
        """
        Object constructor.

        :param unit_length_final: The unit of length used in the generated OpenSCAD code.
        :param fa: The minimum angle (in degrees) of each fragment. Known in OpenSCAD as $fa,
                   see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa.
        :param fs: The minimum circumferential length of each fragment. Known in OpenSCAD as $fs,
                   see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fs.
        :param fn: The number of fragments in 360 degrees. Values of 3 or more override $fa and $fs. Known in OpenSCAD
                   as $fn, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fn.
        :param vpr: The viewport rotation,
                    see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpr.
        :param vpt: The viewport translation,
                    see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpt.
        :param vpf: The FOV (Field of View) of the view,
                    see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpf.
        :param vpd: The camera distance,
                    see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpd.
        :param eps: Epsilon value for clear overlap.
        :param delta: The minimum distance between nodes, vertices and line segments for reliable computation of the
                      separation between line segments and nodes.
        :param angle_digits: The number of decimal places of an angle in the generated OpenSCAD code.
        :param length_digits: The number of decimal places of a length in the generated OpenSCAD code.
        :param scale_digits:  The number of decimal places of a scale or factor in the generated OpenSCAD code.
        """
        self.__project_home: Path = Path(os.getcwd()).resolve()
        """
        The home folder of the current project. 
        """

        self.__target_path: Path | None = None
        """
        The path to the OpenSCAD script that currently been generated.
        """

        self.__code_store: ScadCodeStore = ScadCodeStore()
        """
        The place were we store the generated OpenSCAD code.
        """

        self.__fa: float = fa
        """
        The minimum angle (in degrees) of each fragment. 
        Known in OpenSCAD as $fa, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa.
        """

        self.__fs: float = fs
        """
        The minimum circumferential length of each fragment.
        Known in OpenSCAD as $fs, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fs.
        """

        self.__fn: int = fn
        """
        The number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        Known in OpenSCAD as $fn, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fn.
        """

        self.__vpt: Vector3 = vpt
        """
        The viewport rotation.
        """

        self.__vpr: Vector3 = vpr
        """
        The viewport translation.
        """

        self.__vpd: float = vpd
        """
        The camera distance.
        """

        self.__vpf: float = vpf
        """
        The FOV (Field of View) of the view.
        """

        self.__eps: float = eps
        """
        Epsilon value for clear overlap.
        """

        self.__delta: float = delta
        """
        The minimum distance between nodes, vertices and line segments for reliable computation of the separation 
        between line segments and nodes.
        """

        self.__unit_length_final: Unit = unit_length_final
        """
        The unit of length.
        """

        self.__angle_digits = angle_digits
        """
        The number of decimal places of an angle in the generated OpenSCAD code.
        """

        self.__length_digits = length_digits
        """
        The number of decimal places of a length in the generated OpenSCAD code.
        """

        self.__scale_digits = scale_digits
        """
        The number of decimal places of a scale or factor in the generated OpenSCAD code.
        """

        self.__bom: BOM = BOM()
        """
        The Bill of materials.
        """

        Context.set_unit_length_current(unit_length_final)
        Context.__set_unit_length_final(unit_length_final)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def project_home(self) -> Path:
        """
        Returns the current project's home directory.
        """
        return self.__project_home

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def target_path(self) -> Path | None:
        """
        Returns the path to the OpenSCAD script that currently been generated.
        """
        return self.__target_path

    # ------------------------------------------------------------------------------------------------------------------
    @target_path.setter
    def target_path(self, target_path: str) -> None:
        """
        Set the path to the OpenSCAD script that currently been generated.
        """
        self.__target_path = Path(os.path.realpath(target_path))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_digits(self) -> int:
        """
        Returns the number of decimal places of an angle in the generated OpenSCAD code.
        """
        return self.__angle_digits

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def length_digits(self) -> int:
        """
        Returns the number of decimal places of a length in the generated OpenSCAD code.
        """
        return self.__length_digits

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def scale_digits(self) -> int:
        """
        Returns the number of decimal places of a scale or factor in the generated OpenSCAD code.
        """
        return self.__scale_digits

    # ------------------------------------------------------------------------------------------------------------------
    def resolve_path(self, path: Path | str) -> Path:
        """
        Resolve a path relative from the caller script to a path relative to the project home.

        :param Path path: The path to resolve.
        """
        caller = Path(inspect.stack()[1].filename)
        absolute_path = Path(caller.parent.joinpath(path).resolve())

        if os.path.commonprefix([absolute_path, self.__project_home]) == str(self.__project_home):
            # works with python >=3.12 return absolute_path.relative_to(self.target_path.parent, walk_up=True)
            return Path(os.path.relpath(absolute_path, self.target_path.parent))

        return absolute_path

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def code_store(self) -> ScadCodeStore:
        """
        Returns code store.
        """
        return self.__code_store

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def eps(self) -> float:
        """
        Returns the epsilon value for clear overlap.
        """
        return Length.convert(self.__eps, self.__unit_length_final, self.__unit_length_current)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def delta(self) -> float:
        """
        Returns the minimum distance between nodes, vertices and line segments for reliable computation of the
        separation between line segments and nodes.
        """
        return Length.convert(self.__delta, self.__unit_length_final, self.__unit_length_current)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def resolution(self) -> float:
        """
        Returns the resolution of lengths in generated OpenSCAD code.
        """
        return 10.0 ** -self.__length_digits

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float:
        """
        Returns the minimum angle (in degrees) of each fragment.
        Known in OpenSCAD as $fa, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa.
        """
        return self.__fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float:
        """
        Returns the minimum circumferential length of each fragment.
        Known in OpenSCAD as $fs, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fs.
        """
        return Length.convert(self.__fs, self.__unit_length_final, self.__unit_length_current)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int:
        """
        Returns the number of fragments in 360 degrees. Values of three or more override $fa and $fs.
        Known in OpenSCAD as $fn, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fn.
        """
        return self.__fn

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vpr(self) -> Vector3:
        """
        Returns the viewport rotation,
        see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpr.
        """
        return Vector3(Length.convert(self.__vpr.x, self.__unit_length_final, self.__unit_length_current),
                       Length.convert(self.__vpr.y, self.__unit_length_final, self.__unit_length_current),
                       Length.convert(self.__vpr.z, self.__unit_length_final, self.__unit_length_current))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vpt(self) -> Vector3:
        """
        Returns the viewport translation,
        see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpt.
        """
        return Vector3(Length.convert(self.__vpt.x, self.__unit_length_final, self.__unit_length_current),
                       Length.convert(self.__vpt.y, self.__unit_length_final, self.__unit_length_current),
                       Length.convert(self.__vpt.z, self.__unit_length_final, self.__unit_length_current))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vpf(self) -> float:
        """
        Returns the FOV (Field of View) of the view,
        see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpf.
        """
        return self.__vpf

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vpd(self) -> float:
        """
        Returns the camera distance,
        see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$vpd.
        """
        return Length.convert(self.__vpd, self.__unit_length_final, self.__unit_length_current)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def bom(self) -> BOM:
        """
        Returns the BOM (Bill of Materials).
        """
        return self.__bom

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __set_unit_length_final(__unit_length_final: Unit) -> None:
        """
        Sets the unit of length used in the generated OpenSCAD code.
        """
        Context.__unit_length_final = __unit_length_final

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_unit_length_final() -> Unit:
        """
        Returns the unit of length used in the generated OpenSCAD code.
        """
        return Context.__unit_length_final

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def set_unit_length_current(unit_length_current: Unit) -> None:
        """
        Sets the current unit of length.

        :param unit_length_current: The new current unit of length.
        """
        Context.__unit_length_current = unit_length_current

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_unit_length_current() -> Unit:
        """
        Returns the current unit of length.
        """
        return Context.__unit_length_current

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, List, Set

from super_scad.d2.PolygonMixin import PolygonMixin
from super_scad.d2.private.PrivatePolygon import PrivatePolygon
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2


class Polygon(PolygonMixin, ScadWidget):
    """
    Widget for creating polygons. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#polygon.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 primary: List[Vector2] | None = None,
                 points: List[Vector2] | None = None,
                 secondary: List[Vector2] | None = None,
                 secondaries: List[List[Vector2]] | None = None,
                 convexity: int | None = None,
                 extend_by_eps_sides: bool | List[bool] | Set[int] | None = None):
        """
        Object constructor.

        :param primary: The list of 2D points of the polygon.
        :param points: Alias for primary.
        :param secondary: The secondary path that will be subtracted from the polygon.
        :param secondaries: The secondary paths that will be subtracted form the polygon.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line
                          through the child widget.
        :param extend_by_eps_sides: Whether to extend sides by eps for a clear overlap.
        """
        ScadWidget.__init__(self)
        PolygonMixin.__init__(self, convexity=convexity, extend_by_eps_sides=extend_by_eps_sides)

        self._primary: List[Vector2] | None = primary
        """
        The list of 2D points of the polygon.
        """

        self._points: List[Vector2] | None = points
        """
        Alias for primary.
        """

        self._secondary: List[Vector2] | None = secondary
        """
        The secondary path that will be subtracted from the polygon.
        """

        self._secondaries: List[List[Vector2]] | None = secondaries
        """
        The secondary paths that will be subtracted form the polygon.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the
         constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'primary'}, {'points'})
        validator.validate_exclusive({'secondary'}, {'secondaries'})
        validator.validate_required({'primary', 'points'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def primary(self) -> List[Vector2]:
        """
        Returns the points of the polygon.
        """
        if self._primary is None:
            self._primary = self._points

        return self._primary

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def secondaries(self) -> List[List[Vector2]] | None:
        """
        Returns the points of the polygon.
        """
        if self._secondaries is not None:
            return self._secondaries

        if self._secondary is not None:
            self._secondaries = [self._secondary]

        return self._secondaries

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def nodes(self) -> List[Vector2]:
        """
        Returns the nodes of the polygon.
        """
        return self.primary

    # ------------------------------------------------------------------------------------------------------------------
    def _build_polygon(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        secondaries = self.secondaries
        if secondaries is None:
            return PrivatePolygon(points=self.primary, convexity=self.convexity)

        points = self.primary
        n = 0
        m = n + len(points)
        paths = [list(range(n, m))]
        n = m

        for secondary in secondaries:
            m = n + len(secondary)
            points += secondary
            paths.append(list(range(n, m)))
            n = m

        return PrivatePolygon(points=points, paths=paths, convexity=self.convexity)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.d3.private.PrivateLinearExtrude import PrivateLinearExtrude
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Translate3D import Translate3D
from super_scad.type.Vector2 import Vector2


class LinearExtrude(ScadSingleChildParent):
    """
    Linear Extrusion is an operation that takes a 2D object as input and generates a 3D object as a result. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#linear_extrude.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float,
                 center: bool = False,
                 convexity: int | None = None,
                 twist: float = 0.0,
                 scale: float | Vector2 = 1.0,
                 slices: int | None = None,
                 segments: int | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 extend_by_eps_top: bool = False,
                 extend_by_eps_bottom: bool = False,
                 child: ScadWidget):
        """
        Object constructor.

        :param height: The height of the extruded object.
        :param center: Whether the extruded object is centered along the z-as.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        :param twist: The number of degrees through which the shape is extruded. Setting the parameter twist = 360
                      extrudes through one revolution. The twist direction follows the left-hand rule.
        :param scale: Scales the 2D shape by this value over the height of the extrusion.
        :param slices: Defines the number of intermediate points along the Z axis of the extrusion. Its default
                       increases with the value of twist. Explicitly setting slices may improve the output refinement.
        :param segments: Adds vertices (points) to the extruded polygon resulting in smoother twisted geometries.
                         Segments need to be a multiple of the polygon's fragments to have an effect (6 or 9... for a
                         circle($fn=3), 8,12... for a square()).
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        :param extend_by_eps_top: Whether to extend the top by eps for a clear overlap.
        :param extend_by_eps_bottom: Whether to extend the bottom by eps for a clear overlap.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._height: float = height
        """
        The height of the extruded object.
        """

        self._center: bool = center
        """
        Whether the extruded object is centered along the z-as.
        """

        self._convexity: int | None = convexity
        """
        Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the child
        widget.
        """

        self._twist: float = twist
        """
        The number of degrees through which the shape is extruded. 
        """

        self._scale: float | Vector2 = scale
        """
        Scales the 2D shape by this value over the height of the extrusion.
        """

        self._slices: int | None = slices
        """
        DDefines the number of intermediate points along the Z axis of the extrusion. 
        """

        self._segments: int | None = segments
        """
        Adds vertices (points) to the extruded polygon resulting in smoother twisted geometries.
        """

        self._fa: float | None = fa
        """
        The minimum angle (in degrees) of each fragment.
        """

        self._fs: float | None = fs
        """
        The minimum circumferential length of each fragment.
        """

        self._fn: int | None = fn
        """
        The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        """

        self._extend_by_eps_top: bool = extend_by_eps_top
        """
        Whether to extend the top by eps for a clear overlap.  
        """

        self._extend_by_eps_bottom: bool = extend_by_eps_bottom
        """
        Whether to extend the bottom by eps for a clear overlap.  
        """

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height of the extruded object.
        """
        return self._height

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the extruded object is centered along the z-as.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the
        child widget.
        """
        return self._convexity

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def twist(self) -> float:
        """
        Returns the number of degrees of through which the shape is extruded. Setting the parameter twist = 360
        extrudes through one revolution. The twist direction follows the left-hand rule.
        """
        return self._twist

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def scale(self) -> float | Vector2:
        """
        Returns the number of degrees of through which the shape is extruded. Setting the parameter twist = 360
        extrudes through one revolution. The twist direction follows the left-hand rule.
        """
        return self._scale

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def slices(self) -> int | None:
        """
        Returns the number of intermediate points along the Z axis of the extrusion. Its default
        increases with the value of twist. Explicitly setting slices may improve the output refinement.
        """
        return self._slices

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def segments(self) -> int | None:
        """
        Returns the Adds vertices (points) to the extruded polygon resulting in smoother twisted geometries.
        """
        return self._segments

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._fs

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._fn

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_top(self) -> bool:
        """
        Returns whether to extend the top by eps for a clear overlap.
        """
        return self._extend_by_eps_top

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_bottom(self) -> bool:
        """
        Returns whether to extend the bottom by eps for a clear overlap.
        """
        return self._extend_by_eps_bottom

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        height = self.height
        if self.extend_by_eps_top:
            height += context.eps
        if self.extend_by_eps_bottom:
            height += context.eps

        if (not self.extend_by_eps_bottom and not self.extend_by_eps_top) or \
                (self.extend_by_eps_bottom and self.extend_by_eps_top and self.center):
            return PrivateLinearExtrude(height=height,
                                        center=self.center,
                                        convexity=self.convexity,
                                        twist=self.twist,
                                        scale=self.scale,
                                        slices=self.slices,
                                        segments=self.segments,
                                        fa=self.fa,
                                        fs=self.fs,
                                        fn=self.fn,
                                        child=self.child)

        offset = 0.0
        if self.center:
            offset -= 0.5 * self.height
        if self.extend_by_eps_bottom:
            offset -= context.eps

        return Translate3D(z=offset,
                           child=PrivateLinearExtrude(height=height,
                                                      center=False,
                                                      convexity=self.convexity,
                                                      twist=self.twist,
                                                      scale=self.scale,
                                                      slices=self.slices,
                                                      segments=self.segments,
                                                      fa=self.fa,
                                                      fs=self.fs,
                                                      fn=self.fn,
                                                      child=self.child))

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict, Set

from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.util.Formatter import Formatter


class PrivateOpenScadCommand(ScadWidget):
    """
    Widget for creating OpenSCAD commands.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, command: str, args: Dict[str, Any]):
        """
        Object constructor.

        :param command: The name of the OpenSCAD command.
        :param args: The arguments of the OpenSCAD command.
        """
        ScadWidget.__init__(self)

        self._command: str = command
        """
        The name of the OpenSCAD command.
        """

        self._args: Dict[str, Any] = {}
        """
        The arguments of this OpenSCAD widget.
        """

        if args is not None:
            for key, value in args.items():
                if value is not None and value != self and key not in ('child', 'children'):
                    self._args[key] = value

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return self

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def command(self) -> str:
        """
        Returns the name of the OpenSCAD command.
        """
        return self._command

    # ------------------------------------------------------------------------------------------------------------------
    def generate_args(self, context: Context) -> str:
        """
        Returns the arguments of the OpenSCAD command.
        """
        argument_map = self._argument_map()
        argument_angles = self._argument_angles()
        argument_lengths = self._argument_lengths()
        argument_scales = self._argument_scales()

        args_as_str = '('
        first = True
        for key, value in self._args.items():
            if not first:
                args_as_str += ', '
            else:
                first = False

            real_name = argument_map.get(key, key)
            if real_name in argument_angles:
                real_value = Formatter.format(context, value, is_angle=True)
            elif real_name in argument_lengths:
                real_value = Formatter.format(context, value, is_length=True, unit=self.unit)
            elif real_name in argument_scales:
                real_value = Formatter.format(context, value, is_scale=True)
            else:
                real_value = Formatter.format(context, value)

            if real_name is None:
                args_as_str += '{}'.format(real_value)
            else:
                args_as_str += '{} = {}'.format(real_name, real_value)
        args_as_str += ')'

        return args_as_str

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str | None]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return set()

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return set()

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_scales(self) -> Set[str]:
        """
        Returns the set with arguments that are scales and factors.
        """
        return set()

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand


class Echo(PrivateOpenScadCommand):
    """
    The echo() module prints the contents to the compilation window (aka Console). See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#Echo_module.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, message: str | None = None, **kwargs):
        """
        Object constructor.
        """
        args = {'message': message}
        args.update(kwargs)
        PrivateOpenScadCommand.__init__(self, command='echo', args=args)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str | None]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'message': None}

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Iterable, List, Tuple

from super_scad.boolean.Union import Union
from super_scad.d3.Cylinder import Cylinder
from super_scad.d3.private.PrivatePolyhedron import PrivatePolyhedron
from super_scad.d3.Sphere import Sphere
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Paint import Paint
from super_scad.transformation.Translate3D import Translate3D
from super_scad.type.Color import Color
from super_scad.type.Vector3 import Vector3


# ----------------------------------------------------------------------------------------------------------------------

class Node:
    """
    A node of a polyhedron.
    """

    next_node_id: int = 0
    """
    The next node ID.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, point: Vector3):
        """
        Object constructor.

        :param point: The rounded position of the node.
        """

        self.point: Vector3 = point
        """
        The rounded position of this node.
        """

        self.node_id: int = Node.next_node_id
        """
        The ID of this node as implicitly implied by the invoker.
        """

        Node.next_node_id += 1


# ----------------------------------------------------------------------------------------------------------------------
class Polyhedron(ScadWidget):
    """
    Widget for creating polyhedrons. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#polyhedron.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 faces: List[List[Vector3] | Tuple[Vector3, ...]],
                 highlight_faces: int | List[int] | None = None,
                 highlight_nodes: int | List[int] = None,
                 highlight_diameter: float | None = None,
                 convexity: int | None = None,
                 validate: bool = False,
                 highlight_issues: bool = False):
        """
        Object constructor.

        :param faces: The faces that collectively enclose the solid.
        :param highlight_faces: The ID of the faces to highlight. Each node of the face is marked, the first point
                                is colored red, the second orange, the third green, and all other points are color
                                black.
        :param highlight_nodes: The IDs of the nodes to highlight.
        :param highlight_diameter: The diameter of the spheres that highlight the nodes of the faces.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line
                          the polyhedron.
        :param validate: Whether to validate polyhedron.
        :param highlight_issues: Whether to highlight all issues found by the validation. This will override
                                 highlight_faces and highlight_nodes.

        Each face consists of three or more nodes. Faces may be defined in any order, but the nodes of each face must be
        ordered correctly, must be ordered in clockwise direction when looking at each face from the outside inwards.
        Define enough faces to fully enclose the solid, with no overlap. If nodes that describe a single face are not
        on the same plane, the face is by OpenSCAD automatically split into triangles as needed.
        """
        ScadWidget.__init__(self)

        self._faces: List[List[Vector3] | Tuple[Vector3, ...]] = faces
        """
        The faces that collectively enclose the solid.
        """

        self._highlight_faces: int | List[int] | None = highlight_faces
        """
        The ID of the faces to highlight. Each node of the face is marked, the first point
        """

        self._highlight_nodes: int | List[int] = highlight_nodes
        """
        The IDs of the nodes to highlight.
        """

        self._highlight_diameter: float | None = highlight_diameter
        """
        The diameter of the spheres that highlight the nodes of the faces.
        """

        self._convexity: int | None = convexity
        """
        Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line the polyhedron.
        """

        self._validate: bool = validate
        """
        Whether to validate the polyhedron.
        """

        self._highlight_issues: bool = highlight_issues

        self.__nodes: Dict[int, Node] = {}
        """
        Look up table from the memory address of a position of a node to the node.
        """

        self.__faces: List[List[Node]] = []
        """
        The faces of the polyhedron.
        """

        self.__is_ready: bool = False
        """
        Whether the real faces and the real nodes have been computed. 
        """

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def highlight_faces(self) -> List[int]:
        """
        Returns the IDs of the faces to highlight
        """
        if not isinstance(self._highlight_faces, list):
            if self._highlight_faces is None:
                self._highlight_faces = []

            elif isinstance(self._highlight_faces, int):
                self._highlight_faces = [self._highlight_faces]

        return self._highlight_faces

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def highlight_nodes(self) -> List[int]:
        """
        Returns the IDs of the nodes to highlight
        """
        if not isinstance(self._highlight_nodes, list):
            if self._highlight_nodes is None:
                self._highlight_nodes = []
            elif isinstance(self._highlight_nodes, int):
                self._highlight_nodes = [self._highlight_nodes]

        return self._highlight_nodes

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def highlight_diameter(self) -> float | None:
        """
        Returns the diameter of the spheres that highlight the nodes.
        """
        return self._highlight_diameter

    # ------------------------------------------------------------------------------------------------------------------
    def __real_highlight_diameter(self, context: Context) -> float:
        """
        Returns the real diameter of the spheres that highlight the nodes.
        """
        diameter = self.highlight_diameter
        if diameter is not None:
            return max(diameter, 5.0 * context.resolution)

        face = self._faces[self.highlight_faces[0]]

        total_distance = 0.0
        prev_point = None
        for point in face:
            if prev_point is not None:
                total_distance += (point - prev_point).length
            prev_point = point

        if prev_point is not None:
            total_distance += (face[0] - prev_point).length

        average_distance = total_distance / (len(face) + 1)
        diameter = 0.1 * average_distance

        return round(max(diameter, 5.0 * context.resolution), context.length_digits)

    # ------------------------------------------------------------------------------------------------------------------
    def nodes(self, context: Context) -> List[Vector3]:
        """
        Returns the nodes of the polyhedron.
        """
        self.__prepare_data(context)

        return [node.point for node in self.__nodes.values()]

    # ------------------------------------------------------------------------------------------------------------------
    def faces(self, context: Context) -> List[List[int]]:
        """
        Returns the real faces of the polyhedron.
        """
        self.__prepare_data(context)

        return [[node.node_id for node in face] for face in self.__faces]

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the
        child widget.
        """
        return self._convexity

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def validate(self) -> bool:
        """
        Returns whether to validate polyhedron.
        """
        return self._validate

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def highlight_issues(self) -> bool:
        """
        Returns whether to highlight all issues found by the validation.
        """
        return self._highlight_issues

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __create_face_marker_nodes(points: Iterable[Vector3], is_real_face: bool, diameter: float) -> List[ScadWidget]:
        """
        Creates markers of the nodes of a face.

        :param points: The face, given as an iterable over its nodes given as points.
        :param is_real_face: Whether the face is a real face.
        :param diameter: The diameter of the markers.
        """
        markers = []
        for node_id, node in enumerate(points):
            if is_real_face:
                if node_id == 0:
                    color = Color('red')
                elif node_id == 1:
                    color = Color('orange')
                elif node_id == 2:
                    color = Color('green')
                else:
                    color = Color('black')
            else:
                color = Color('pink')

            marker = Paint(color=color,
                           child=Translate3D(vector=node,
                                             child=Sphere(diameter=diameter, fn=16)))
            markers.append(marker)

        return markers

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __create_face_marker_edges(points: Iterable[Vector3],
                                   is_real_face: bool,
                                   diameter: float,
                                   context: Context) -> List[ScadWidget]:
        """
        Creates markers of the edges of a face.

        :param points: The face, given as an iterable over its nodes given as points.
        :param is_real_face: Whether the face is a real face.
        :param diameter: The diameter of cylinders on the edges.
        :param context: The build context.
        """
        if is_real_face:
            color = Color('black')
        else:
            color = Color('pink')

        edges = []
        prev_point = None
        first_point = None
        for point in points:
            if prev_point is None:
                first_point = point
            else:
                if (point - prev_point).length >= context.resolution:
                    edge = Paint(color=color,
                                 child=Cylinder(start_point=prev_point,
                                                end_point=point,
                                                diameter=diameter,
                                                fn=16))
                    edges.append(edge)

            prev_point = point

        if prev_point is not None and first_point is not None:
            if (first_point - prev_point).length >= context.resolution:
                edge = Paint(color=color,
                             child=Cylinder(start_point=prev_point,
                                            end_point=first_point,
                                            diameter=diameter))
                edges.append(edge)

        return edges

    # ------------------------------------------------------------------------------------------------------------------
    def __create_markers(self, context: Context) -> List[ScadWidget]:
        """
        Create markers to faces and points.

        :param context: The build context.
        """
        markers = self.__create_markers_faces(context)
        markers += self.__create_markers_node(context)

        return markers

    # ------------------------------------------------------------------------------------------------------------------
    def __create_markers_faces(self, context: Context) -> List[ScadWidget]:
        """
        Create markers to highlight a face.

        :param context: The build context.
        """
        diameter_node = self.__real_highlight_diameter(context)
        diameter_edge = 0.2 * diameter_node
        markers = []

        for face_id in self.highlight_faces:
            if 0 <= face_id < len(self.__faces):
                face = self.__faces[face_id]

                points = [node.point for node in face]
                markers += self.__create_face_marker_nodes(points, True, diameter_node)
                markers += self.__create_face_marker_edges(points, True, diameter_edge, context)

                node_ids = ', '.join(str(node.node_id) for node in face)
                print(f'Note: Highlighting face {face_id}: [{node_ids}].')
                for node in face:
                    print(f'      Node {node.node_id}: {node.point}.')
            else:
                print(f'Warning: Face {face_id} not found.')

        return markers

    # ------------------------------------------------------------------------------------------------------------------
    def __create_markers_node(self, context: Context) -> List[ScadWidget]:
        """
        Create markers to highlight a node.

        :param context: The build context.
        """
        diameter_node = self.__real_highlight_diameter(context)
        markers = []

        map_node_id_to_node = {}
        for node in self.__nodes.values():
            map_node_id_to_node[node.node_id] = node

        for node_id in self.highlight_nodes:
            if node_id in map_node_id_to_node:
                point = map_node_id_to_node[node_id].point
                markers.append(Paint(color=Color('blue'),
                                     child=Translate3D(vector=point,
                                                       child=Sphere(diameter=diameter_node, fn=16))))

                print(f'Note: Highlighting node {node_id}: {point}.')
            else:
                print(f'Warning: Node {node_id} not found.')

        return markers

    # ------------------------------------------------------------------------------------------------------------------
    def __prepare_data(self, context):
        """
        Prepares the data as expected by OpenSCAD polyhedron.

        @param context: The build context.
        """
        if not self.__is_ready:
            Node.next_node_id = 0
            digits = context.length_digits

            for points in self._faces:
                face = []
                for point in points:
                    node = self.__nodes.get(id(point))
                    if node is None:
                        point_rounded = Vector3(round(point.x, digits),
                                                round(point.y, digits),
                                                round(point.z, digits))
                        node = Node(point_rounded)
                        self.__nodes[id(point)] = node
                    face.append(node)
                self.__faces.append(face)

            self.__is_ready = True

    # ------------------------------------------------------------------------------------------------------------------
    def __validate_line_segments(self) -> None:
        """
        Validates that all line segments are connected to exactly two faces.
        """
        line_segment_to_faces_map: Dict[Tuple[int, int], List[int]] = {}
        for face_id, face in enumerate(self.__faces):
            length = len(face)
            for index in range(length):
                node_id1 = face[index].node_id
                node_id2 = face[(index + 1) % length].node_id
                line_segment = (min(node_id1, node_id2), max(node_id1, node_id2))
                if line_segment in line_segment_to_faces_map:
                    line_segment_to_faces_map[line_segment].append(face_id)
                else:
                    line_segment_to_faces_map[line_segment] = [face_id]

        for line_segment, original_face_ids in line_segment_to_faces_map.items():
            if len(original_face_ids) != 2:
                face_ids = ', '.join(str(original_face_id) for original_face_id in original_face_ids)

                if len(original_face_ids) > 2:
                    print(f'Warning: Line segment {line_segment} is part of one face only.')
                    print(f'         Add highlight_nodes=[{line_segment}] to locate this line segment.')
                    print(f'         Add highlight_faces=[{face_ids}] to locate the face.')

                elif len(original_face_ids) < 2:
                    numer_of_faces = len(original_face_ids)
                    print(f'Warning: Line segment {line_segment} is part of {numer_of_faces} faces.')
                    print(f'         Add highlight_nodes=[{line_segment}] to locate this line segment.')
                    print(f'         Add highlight_faces=[{face_ids}] to locate the faces.')

                if self.highlight_issues:
                    self._highlight_nodes.append(line_segment[0])
                    self._highlight_nodes.append(line_segment[1])

    # ------------------------------------------------------------------------------------------------------------------
    def __validate_faces1(self) -> None:
        """
        Validates that faces have three or more nodes.
        """
        for face_id, face in enumerate(self.__faces):
            length = len(face)
            if length < 3:
                print(f'Warning: Face has less than three nodes.')
                print(f'         Add highlight_faces=[{face_id}] to locate the face.')

                if self.highlight_issues:
                    self._highlight_faces.append(face_id)

    # ------------------------------------------------------------------------------------------------------------------
    def __validate_faces2(self) -> None:
        """
        Validates that the nodes of all faces are ordered in clockwise direction.
        """
        # For now, press F12.
        pass

    # ------------------------------------------------------------------------------------------------------------------
    def __validate(self) -> None:
        """
        Validates the polyhedron.
        """
        if self.highlight_issues:
            self._highlight_faces = []
            self._highlight_nodes = []

        self.__validate_line_segments()
        self.__validate_faces1()
        self.__validate_faces2()

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        self.__prepare_data(context)
        if self.validate:
            self.__validate()

        polyhedron = PrivatePolyhedron(points=self.nodes(context),
                                       faces=self.faces(context),
                                       convexity=self.convexity)

        if not self.highlight_faces and not self.highlight_nodes:
            return polyhedron

        markers = self.__create_markers(context)

        return Union(children=[polyhedron] + markers)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.d2.Semicircle import Semicircle
from super_scad.d3.private.PrivateSphere import PrivateSphere
from super_scad.d3.RotateExtrude import RotateExtrude
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Rotate2D import Rotate2D
from super_scad.util.Radius2Sides4n import Radius2Sides4n


class Sphere(ScadWidget):
    """
    Class for spheres. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#sphere.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 radius: float | None = None,
                 diameter: float | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 fn4n: bool | None = None):
        """
        Object constructor.

        :param radius: The radius of the sphere.
        :param diameter: The diameter of the sphere.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of three or more override fa and fs.
        :param fn4n: Whether to create a sphere with a multiple of four vertices.
        """
        ScadWidget.__init__(self)

        self._radius: float | None = radius
        """
        The radius of the sphere.
        """

        self._diameter: float | None = diameter
        """
        The diameter of the sphere.
        """

        self._fa: float | None = fa
        """
        The minimum angle (in degrees) of each fragment.
        """

        self._fs: float | None = fs
        """
        The minimum circumferential length of each fragment.
        """

        self._fn: int | None = fn
        """
        The fixed number of fragments in 360 degrees.
        """

        self._fn4n: bool | None = fn4n
        """
        Whether to create a sphere with a multiple of four vertices.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'radius'}, {'diameter'})
        validator.validate_required({'radius', 'diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def radius(self) -> float:
        """
        Returns the radius of the sphere.
        """
        if self._radius is None:
            self._radius = 0.5 * self._diameter

        return self._radius

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def diameter(self) -> float:
        """
        Returns the diameter of the sphere.
        """
        if self._diameter is None:
            self._diameter = 2.0 * self._radius

        return self._diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._fs

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of three or more override fa and fs.
        """
        return self._fn

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn4n(self) -> bool:
        """
        Returns whether to create a circle with multiple of 4 vertices.
        """
        if self._fn4n is None:
            self._fn4n = False

        return self._fn4n

    # ------------------------------------------------------------------------------------------------------------------
    def real_fn(self, context: Context) -> int | None:
        """
        Returns the real fixed number of fragments in 360 degrees.
        """
        if self.fn4n:
            return Radius2Sides4n.r2sides4n(context, self.radius)

        return self.fn

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        if not self.fn4n:
            return PrivateSphere(diameter=self.diameter, fa=self.fa, fs=self.fs, fn=self.real_fn(context))

        return RotateExtrude(angle=360.0,
                             fn=self.real_fn(context),
                             child=Rotate2D(angle=-90.0, child=Semicircle(diameter=self.diameter, fn4n=True)))

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.d3.Cuboid import Cuboid
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type import Vector3


class Cube(ScadWidget):
    """
    Widget for creating cubes. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#cube.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: float,
                 center: bool = False,
                 extend_by_eps_front: bool = False,
                 extend_by_eps_back: bool = False,
                 extend_by_eps_top: bool = False,
                 extend_by_eps_bottom: bool = False,
                 extend_by_eps_left: bool = False,
                 extend_by_eps_right: bool = False):
        """
        Object constructor.

        :param size: The size of the cube.
        :param center: Whether the cube is centered at the origin.
        :param extend_by_eps_front: Whether to extend the front face of the cube by eps for a clear overlap.
        :param extend_by_eps_back: Whether to extend the back face of the cube by eps for a clear overlap.
        :param extend_by_eps_top: Whether to extend the top face of the cube by eps for a clear overlap.
        :param extend_by_eps_bottom: Whether to extend the bottom face of the cube by eps for a clear overlap.
        :param extend_by_eps_left: Whether to extend the left face of the cube by eps for a clear overlap.
        :param extend_by_eps_right: Whether to extend the right face of the cube by eps for a clear overlap.
        """
        ScadWidget.__init__(self)

        self._size: float = size
        """
        The size of the cube.
        """

        self._center: bool = center
        """
        Whether the cube is centered at the origin.
        """

        self._extend_by_eps_front: bool = extend_by_eps_front
        """
        Whether to extend the front face of the cube by eps for a clear overlap.
        """

        self._extend_by_eps_back: bool = extend_by_eps_back
        """        
        Whether to extend the back face of the cube by eps for a clear overlap.
        """

        self._extend_by_eps_top: bool = extend_by_eps_top
        """
        Whether to extend the top face of the cube by eps for a clear overlap.
        """

        self._extend_by_eps_bottom: bool = extend_by_eps_bottom
        """
        Whether to extend the bottom face of the cube by eps for a clear overlap.
        """

        self._extend_by_eps_left: bool = extend_by_eps_left
        """
        Whether to extend the left face of the cube by eps for a clear overlap.
        """

        self._extend_by_eps_right: bool = extend_by_eps_right
        """
        Whether to extend the right face of the cube by eps for a clear overlap.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_required({'size'}, {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cube is centered at the origin.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> float:
        """
        Returns the size of the cube.
        """
        return self._size

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_front(self) -> bool:
        """
        Return whether to extend the front face of the cube by eps for a clear overlap.
        """
        return self._extend_by_eps_front

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_back(self) -> bool:
        """
        Return whether to extend the back face of the cube by eps for a clear overlap.
        """
        return self._extend_by_eps_back

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_top(self) -> bool:
        """
        Return whether to extend the top face of the cube by eps for a clear overlap.
        """
        return self._extend_by_eps_top

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_bottom(self) -> bool:
        """
        Return whether to extend the bottom face of the cube by eps for a clear overlap.
        """
        return self._extend_by_eps_bottom

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_left(self) -> bool:
        """
        Return whether to extend the left face of the cube by eps for a clear overlap.
        """
        return self._extend_by_eps_left

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_right(self) -> bool:
        """
        Return whether to extend the front face of the cube by eps for a clear overlap.
        """
        return self._extend_by_eps_right

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return Cuboid(size=Vector3(self.size, self.size, self.size),
                      center=self.center,
                      extend_by_eps_front=self.extend_by_eps_front,
                      extend_by_eps_back=self.extend_by_eps_back,
                      extend_by_eps_top=self.extend_by_eps_top,
                      extend_by_eps_bottom=self.extend_by_eps_bottom,
                      extend_by_eps_left=self.extend_by_eps_left,
                      extend_by_eps_right=self.extend_by_eps_right)

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.scad.Unit import Unit
from super_scad.scad.unit.AngstromUnit import AngstromUnit
from super_scad.scad.unit.AstronomicalUnit import AstronomicalUnit
from super_scad.scad.unit.AttoparsecUnit import AttoparsecUnit
from super_scad.scad.unit.CentimeterUnit import CentimeterUnit
from super_scad.scad.unit.DecimeterUnit import DecimeterUnit
from super_scad.scad.unit.FootUnit import FootUnit
from super_scad.scad.unit.InchUnit import InchUnit
from super_scad.scad.unit.KilometerUnit import KilometerUnit
from super_scad.scad.unit.LengthUnit import LengthUnit
from super_scad.scad.unit.LightYearUnit import LightYearUnit
from super_scad.scad.unit.LiUnit import LiUnit
from super_scad.scad.unit.MeterUnit import MeterUnit
from super_scad.scad.unit.MicrometerUnit import MicrometerUnit
from super_scad.scad.unit.MileUnit import MileUnit
from super_scad.scad.unit.MillimeterUnit import MillimeterUnit
from super_scad.scad.unit.ParsecUnit import ParsecUnit
from super_scad.scad.unit.RoyalCubitUnit import RoyalCubitUnit
from super_scad.scad.unit.ThouUnit import ThouUnit
from super_scad.scad.unit.YardUnit import YardUnit


class Length:
    """
    Utility class for converting lengths between different units of length.
    """
    # ------------------------------------------------------------------------------------------------------------------
    __ratio: List[List[float] | None] = []
    """
    The ratios of all unit of lengths.
    """

    __units: List = [AngstromUnit,
                     AstronomicalUnit,
                     AttoparsecUnit,
                     CentimeterUnit,
                     DecimeterUnit,
                     FootUnit,
                     InchUnit,
                     KilometerUnit,
                     LightYearUnit,
                     LiUnit,
                     MeterUnit,
                     MicrometerUnit,
                     MileUnit,
                     MillimeterUnit,
                     ParsecUnit,
                     RoyalCubitUnit,
                     ThouUnit,
                     YardUnit]
    """
    All unit of unit of lengths known to SuperSCAD.
    """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def init_ratios() -> None:
        """
        Initializes the ratios of lengths.
        """
        max_value = max(unit.value for unit in Unit)
        for i in range(max_value + 1):
            for j in range(max_value + 1):
                if j == 0:
                    Length.__ratio.append([])
                Length.__ratio[i].append(None)

        for unit_from_class in Length.__units:
            unit_from = unit_from_class()
            id_from = unit_from.id()
            for unit_to_class in Length.__units:
                unit_to = unit_to_class()
                id_to = unit_to.id()
                Length.__ratio[id_from][id_to] = unit_from.meters() / unit_to.meters()

        # One special case. A length that is free of any scale can be converted to a length that is free of any scale
        # only.
        Length.__ratio[Unit.FREE.value][Unit.FREE.value] = 1.0

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def convert(length: float, from_unit: Unit, to_unit: Unit) -> float:
        """
        Converts a length from one unit to another.

        @param length: The length to convert.
        @param from_unit: The unit to convert from.
        @param to_unit: The unit to convert to.
        """
        try:
            return length * Length.__ratio[from_unit.value][to_unit.value]
        except Exception:
            raise ValueError(f'Cannot convert length {length} from {from_unit.name} to {to_unit.name}.')

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def length_unit(unit: Unit) -> LengthUnit:
        """
        Returns the length unit object of a unit of length.

        :param unit: The unit of length.
        """
        if unit == Unit.UM:
            return MicrometerUnit()

        if unit == Unit.MM:
            return MillimeterUnit()

        if unit == Unit.CM:
            return CentimeterUnit()

        if unit == Unit.DM:
            return DecimeterUnit()

        if unit == Unit.M:
            return MeterUnit()

        if unit == Unit.KM:
            return KilometerUnit()

        if unit == Unit.THOU:
            return ThouUnit()

        if unit == Unit.INCH:
            return InchUnit()

        if unit == Unit.FOOT:
            return FootUnit()

        if unit == Unit.YARD:
            return YardUnit()

        if unit == Unit.MILE:
            return MileUnit()

        if unit == Unit.LI:
            return LiUnit()

        if unit == Unit.ROYAL_CUBIT:
            return RoyalCubitUnit()

        if unit == Unit.ANGSTROM:
            return AngstromUnit()

        if unit == Unit.ASTRONOMICAL_UNIT:
            return AstronomicalUnit()

        if unit == Unit.LIGHT_YEAR:
            return LightYearUnit()

        if unit == Unit.PARSEC:
            return ParsecUnit()

        if unit == Unit.ATTOPARSEC:
            return AttoparsecUnit()

        raise ValueError(f'Unsupported unit {unit}.')

# ----------------------------------------------------------------------------------------------------------------------



from typing import List, Tuple

from super_scad.boolean.Compound import Compound
from super_scad.boolean.Difference import Difference
from super_scad.scad.ScadWidget import ScadWidget


class YinYang:
    """
    A helper class for adding and subtracting widgets.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self):
        """
        Object contructor.
        """
        self._negatives: List[ScadWidget] = []
        self._positives: List[ScadWidget] = []

    # ------------------------------------------------------------------------------------------------------------------
    def __add__(self, other: Tuple[ScadWidget | None, ScadWidget | None]):
        """
        Adds an optional negative and an optional positive widget to the lists of negative and positive widgets.

        :param other: A tuple with an optional negative (the first element of the tuple) and an optional positive
                     (the second element of the tuple) widget.
        """
        if other[0] is not None:
            assert isinstance(other[0], ScadWidget)
            self._negatives.append(other[0])
        if other[1] is not None:
            assert isinstance(other[1], ScadWidget)
            self._positives.append(other[1])

        return self

    # ------------------------------------------------------------------------------------------------------------------
    def apply_negatives_positives(self, body: ScadWidget) -> ScadWidget:
        """
        First, subtracts all negatives from the body and then adds all positives.

        :param body: The body widget.
        """
        if self._negatives:
            body = Difference(children=[body, *self._negatives])
        if self._positives:
            body = Compound(children=[body, *self._positives])

        return body

    # ------------------------------------------------------------------------------------------------------------------
    def apply_positives_negatives(self, body: ScadWidget) -> ScadWidget:
        """
        First, adds all positives to the body and then subtracts all negatives.

        :param body: The body widget.
        """
        if self._positives:
            body = Compound(children=[body, *self._positives])
        if self._negatives:
            body = Difference(children=[body, *self._negatives])

        return body

# ----------------------------------------------------------------------------------------------------------------------



from typing import List, Set


class ArgumentValidator:
    """
    Class for the validation of arguments of a SupeSCAD widget constructor.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, args):
        """
        Object constructor.

        """
        self.__arguments_set: Set[str] = set(name for name, value in args.items() if value is not None)
        """
        A set with all the none empty arguments passed to the constructor of a ScadWidget.
        """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _to_string(sets: List[Set[str] | str]) -> str:
        """

        """
        if len(sets) == 0:
            return ''

        if len(sets) == 1:
            return str(sets[0])

        if len(sets) == 2:
            return str(sets[0]) + ' and ' + str(sets[1])

        ret = ''
        for i in range(len(sets) - 1):
            ret += str(sets[i]) + ', '
        ret += 'and ' + str(sets[-1])

        return ret

    # ------------------------------------------------------------------------------------------------------------------
    def validate_exclusive(self, *args: Set[str] | str) -> None:
        """
        Validates that only arguments belonging to one of the given sets are passed to the constructor of the SuperScad
        widget.
        """
        supplied_sets = []
        for index, arg in enumerate(args):
            if isinstance(arg, str):
                supplied = arg in self.__arguments_set
            elif isinstance(arg, set):
                supplied = set.intersection(self.__arguments_set, arg)
            else:
                raise TypeError(f'Argument {index + 1} must be a set or str, got a {type(arg)} instead.')
            if supplied:
                supplied_sets.append(supplied)

        if len(supplied_sets) > 1:
            sets = self._to_string(supplied_sets)
            raise ValueError(f'The following set of arguments are not exclusive: {sets}.')

    # ------------------------------------------------------------------------------------------------------------------
    def validate_required(self, *args: Set[str] | str) -> None:
        """
        Validates that at least one argument belonging to each of the given sets are passed to the constructor of the
        SuperSCAD widget.
        """
        for index, arg in enumerate(args):
            if isinstance(arg, str):
                supplied = arg in self.__arguments_set
                if not supplied:
                    raise ValueError(f'Argument {arg} is mandatory.')
            elif isinstance(arg, set):
                supplied = set.intersection(self.__arguments_set, arg)
                if not supplied:
                    raise ValueError(f'At least one of these arguments {str(arg)} must be supplied.')
            else:
                raise TypeError(f'Argument {index + 1} must be a set or str, got a {type(arg)} instead.')

    # ------------------------------------------------------------------------------------------------------------------
    def validate_count(self, count: int, *args: Set[str] | str) -> None:
        """
        Validates that exactly the given number of arguments are passed to the constructor of the SuperSCAD widget.
        """
        actual = 0
        for index, arg in enumerate(args):
            if isinstance(arg, str):
                if arg in self.__arguments_set:
                    actual += 1
            elif isinstance(arg, set):
                if set.intersection(self.__arguments_set, arg):
                    actual += 1
            else:
                raise TypeError(f'Argument {index + 1} must be a set or str, got a {type(arg)} instead.')

        if actual != count:
            arguments = self._to_string(list(args))
            raise ValueError(f'Of arguments {arguments} exactly {count} must be supplied, got {actual} instead.')

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.private.PrivateMultiChildOpenScadCommand import PrivateMultiChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class Minkowski(PrivateMultiChildOpenScadCommand):
    """
    Displays the minkowski sum of the child widgets. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#minkowski.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 convexity: int | None = None,
                 children: List[ScadWidget]):
        """
        Object constructor.

        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          the child widgets.
        :param children: The child widgets.
        """
        PrivateMultiChildOpenScadCommand.__init__(self, command='minkowski', args=locals(), children=children)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.boolean.Intersection import Intersection
from super_scad.d2.Circle import Circle
from super_scad.d2.Rectangle import Rectangle
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Translate2D import Translate2D


class Semicircle(ScadWidget):
    """
    Widget for creating semicircles.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 radius: float | None = None,
                 diameter: float | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 fn4n: bool | None = None):
        """
        Object constructor.

        :param radius: The radius of the semicircle.
        :param diameter: The diameter of the semicircle.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of three or more override fa and fs.
        :param fn4n: Whether to create a semicircle based on circle with a multiple of four vertices.
        """
        ScadWidget.__init__(self)

        self._radius: float | None = radius
        """
        The radius of the semicircle.
        """

        self._diameter: float | None = diameter
        """
        The diameter of the semicircle.
        """

        self._fa: float | None = fa
        """
        The minimum angle (in degrees) of each fragment.
        """

        self._fs: float | None = fs
        """
        The fixed number of fragments in 360 degrees.
        """

        self._fn: int | None = fn
        """
        The fixed number of fragments in 360 degrees.
        """

        self._fn4n: bool | None = fn4n
        """
        Whether to create a semicircle based on circle with a multiple of four vertices.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'radius'}, {'diameter'})
        validator.validate_exclusive({'fn4n'}, {'fa', 'fs', 'fn'})
        validator.validate_required({'radius', 'diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def radius(self) -> float:
        """
        Returns the radius of this semicircle.
        """
        if self._radius is None:
            self._radius = 0.5 * self._diameter

        return self._radius

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def diameter(self) -> float:
        """
        Returns the diameter of this semicircle.
        """
        if self._diameter is None:
            self._diameter = 2.0 * self._radius

        return self._diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._fs

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._fn

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn4n(self) -> bool | None:
        """
        Returns whether to create a semicircle with multiple of four vertices.
        """
        return self._fn4n

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return Intersection(children=[Circle(diameter=self.diameter,
                                             fa=self.fa,
                                             fs=self.fs,
                                             fn=self.fn,
                                             fn4n=self.fn4n),
                                      Translate2D(x=-(self.radius + context.eps),
                                                  child=Rectangle(width=self.diameter + 2 * context.eps,
                                                                  depth=self.radius + context.eps))])

# ----------------------------------------------------------------------------------------------------------------------



import math

from super_scad.type import Vector2


class LineIntersection2D:
    """
    A utility class for computing the intersection point of two lines in 2D space.
    """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def intersection(line_a1: Vector2,
                     line_a2: Vector2,
                     line_b1: Vector2,
                     line_b2: Vector2) -> Vector2 | None:
        """
        Returns the intersection point between two lines given tow points on each line.

        :param line_a1: The first point on the first line.
        :param line_a2: The second point on the first line.
        :param line_b1: The first point on the second line.
        :param line_b2: The second point on the second line.
        """
        a1 = line_a2.y - line_a1.y
        b1 = line_a1.x - line_a2.x
        c1 = a1 * line_a1.x + b1 * line_a1.y

        a2 = line_b2.y - line_b1.y
        b2 = line_b1.x - line_b2.x
        c2 = a2 * line_b1.x + b2 * line_b1.y

        determinant = a1 * b2 - a2 * b1

        if determinant == 0.0:
            return None

        x = (b2 * c1 - b1 * c2) / determinant
        y = (a1 * c2 - a2 * c1) / determinant

        if not math.isfinite(x) or not math.isfinite(y):
            return None

        return Vector2(x, y)

# ----------------------------------------------------------------------------------------------------------------------



from pathlib import Path

from super_scad.private.PrivateMultiChildOpenScadCommand import PrivateMultiChildOpenScadCommand
from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.util.Formatter import Formatter


class Scad:
    """
    The SuperSCAD super object for running SuperSCAD.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, context: Context):
        """
        Object constructor.

        :param context: The build context.
        """

        self.__context: Context = context
        """
        The build context.
        """

    # ------------------------------------------------------------------------------------------------------------------
    def run_super_scad(self, root_widget: ScadWidget, openscad_path: Path | str) -> None:
        """
        Runs SuperSCAD on a SuperSCAD widget and stores the generated OpenSCAD code.

        :param root_widget: The root SuperSCAD widget to build.
        :param openscad_path: The path to the file where to store the generated OpenSCAD code.
        """
        self.__run_super_scad_prepare(openscad_path)
        self.__run_super_scad_walk_build_tree(root_widget)
        self.__run_super_scad_finalize()

    # ------------------------------------------------------------------------------------------------------------------
    def __run_super_scad_prepare(self, openscad_path: Path | str) -> None:
        """
        Executes the required steps before running SuperSCAD.

        :param openscad_path: The path to the file where to store the generated OpenSCAD code.
        """
        self.__context.target_path = Path(openscad_path)
        self.__context.set_unit_length_current(self.__context.get_unit_length_final())
        self.__context.code_store.clear()

        self.__context.code_store.add_line('// Unit of length: {}'.format(Context.get_unit_length_final()))

        if self.__context.fa != Context.DEFAULT_FA:
            fa = Formatter.format(self.__context, self.__context.fa, is_angle=True)
            self.__context.code_store.add_line(f'$fa = {fa};')
        if self.__context.fs != Context.DEFAULT_FS:
            fs = Formatter.format(self.__context,
                                  self.__context.fs,
                                  is_length=True,
                                  unit=self.__context.get_unit_length_final())
            self.__context.code_store.add_line(f'$fs = {fs};')
        if self.__context.fn != Context.DEFAULT_FN:
            self.__context.code_store.add_line(f'$fn = {self.__context.fn};')
        if self.__context.vpt != Context.DEFAULT_VIEWPORT_TRANSLATION:
            vpt = Formatter.format(self.__context,
                                   self.__context.vpt,
                                   is_length=True,
                                   unit=self.__context.get_unit_length_final())
            self.__context.code_store.add_line(f'$vpt = {vpt};')
        if self.__context.vpr != Context.DEFAULT_VIEWPORT_ROTATION:
            vpr = Formatter.format(self.__context, self.__context.vpr, is_angle=True)
            self.__context.code_store.add_line(f'$vpr = {vpr};')
        if self.__context.vpd != Context.DEFAULT_VIEWPORT_DISTANCE:
            vpd = Formatter.format(self.__context,
                                   self.__context.vpd,
                                   is_length=True,
                                   unit=self.__context.get_unit_length_final())
            self.__context.code_store.add_line(f'$vpd = {vpd};')
        if self.__context.vpf != Context.DEFAULT_VIEWPORT_FIELD_OF_VIEW:
            vpf = Formatter.format(self.__context, self.__context.vpf, is_angle=True)
            self.__context.code_store.add_line(f'$vpf = {vpf};')

        if self.__context.code_store.line_count() > 1:
            self.__context.code_store.add_line('')

    # ------------------------------------------------------------------------------------------------------------------
    def __run_super_scad_finalize(self) -> None:
        """
        Executes the required step after running SuperSCAD.
        """
        self.__context.code_store.add_line('')

        with open(self.__context.target_path, 'wt') as handle:
            handle.write(self.__context.code_store.get_code())

    # ------------------------------------------------------------------------------------------------------------------
    def __run_super_scad_walk_build_tree(self, parent_widget: ScadWidget) -> None:
        """
        Helper method for __run_super_scad. Runs recursively on the SubSCAD widget and its children until it finds a
        widget for an OpenSCAD command. This OpenSCAD command is used to generate the OpenSCAD code.

        :param parent_widget: The parent widget to build.
        """
        old_unit = Context.get_unit_length_current()
        self.__context.set_unit_length_current(parent_widget.unit)
        child_widget = parent_widget.build(self.__context)
        Context.set_unit_length_current(old_unit)

        if isinstance(child_widget, PrivateOpenScadCommand):
            self.__context.code_store.add_line('{}{}'.format(child_widget.command,
                                                             child_widget.generate_args(self.__context)))

            if isinstance(child_widget, PrivateSingleChildOpenScadCommand):
                self.__context.code_store.add_line('{')
                self.__run_super_scad_walk_build_tree(child_widget.child)
                self.__context.code_store.add_line('}')

            elif isinstance(child_widget, PrivateMultiChildOpenScadCommand):
                self.__context.code_store.add_line('{')
                for child in child_widget.children:
                    self.__run_super_scad_walk_build_tree(child)
                self.__context.code_store.add_line('}')

            else:
                self.__context.code_store.append_to_last_line(';')

        else:
            if child_widget == parent_widget:
                # Only OpenSCAD commands are allowed to build themselves.
                ValueError(f'Widget {parent_widget.__class__} build itself.')

            self.__run_super_scad_walk_build_tree(child_widget)

# ----------------------------------------------------------------------------------------------------------------------



import math
from typing import Any, Dict

from super_scad.d3.private.PrivateCylinder import PrivateCylinder
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Rotate3D import Rotate3D
from super_scad.transformation.Translate3D import Translate3D
from super_scad.type.Vector3 import Vector3
from super_scad.util.Radius2Sides4n import Radius2Sides4n


class Cylinder(ScadWidget):
    """
    Widget for creating cylinders.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float | None = None,
                 start_point: Vector3 | None = None,
                 end_point: Vector3 | None = None,
                 radius: float | None = None,
                 diameter: float | None = None,
                 center: bool | None = None,
                 extend_by_eps_top: bool = False,
                 extend_by_eps_bottom: bool = False,
                 extend_by_eps_radius: bool = False,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 fn4n: bool | None = None):
        """
        Object constructor.

        :param height: The height of the cylinder.
        :param start_point: The start point of the cylinder.
        :param end_point: The end point of the cylinder.
        :param radius: The radius of the cylinder.
        :param diameter: The diameter of the cylinder.
        :param center: Whether the cylinder is centered along the z-as. Defaults to false.
        :param extend_by_eps_top: Whether to extend the top of the cylinder by eps for a clear overlap.
        :param extend_by_eps_bottom: Whether to extend the bottom of the cylinder by eps for a clear overlap.
        :param extend_by_eps_radius: Whether to extend the radius of the cylinder by eps for a clear overlap.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        :param fn4n: Whether to create a cylinder with a multiple of 4 vertices.
        """
        ScadWidget.__init__(self)

        self._height: float | None = height
        """
        The height of the cylinder.
        """

        self._start_point: Vector3 | None = start_point
        """
        The start point of the cylinder.
        """

        self._end_point: Vector3 | None = end_point
        """
        The end point of the cylinder.
        """

        self._radius: float | None = radius
        """
        The radius of the cylinder.
        """

        self._diameter: float | None = diameter
        """
        The diameter of the cylinder.
        """

        self._center: bool | None = center
        """
        Whether the cylinder is centered along the z-as. Defaults to false.
        """

        self._extend_by_eps_top: bool = extend_by_eps_top
        """
        Whether to extend the top of the cylinder by eps for a clear overlap.
        """

        self._extend_by_eps_bottom: bool = extend_by_eps_bottom
        """
        Whether to extend the bottom of the cylinder by eps for a clear overlap.
        """

        self._extend_by_eps_radius: bool = extend_by_eps_radius
        """
        Whether to extend the radius of the cylinder by eps for a clear overlap.
        """

        self._fa: float | None = fa
        """
        The minimum angle (in degrees) of each fragment.
        """

        self._fs: float | None = fs
        """
        The minimum circumferential length of each fragment.
        """

        self._fn: int | None = fn
        """
        The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        """

        self._fn4n: bool | None = fn4n
        """
        Whether to create a cylinder with a multiple of 4 vertices.
        """

        self._explicit_height: bool = height is not None
        """
        Whether the height is explicit specified.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'height'}, {'start_point', 'end_point'})
        validator.validate_exclusive({'radius'}, {'diameter'})
        validator.validate_exclusive({'fn4n'}, {'fa', 'fs', 'fn'})
        validator.validate_required({'height', 'start_point'},
                                    {'height', 'end_point'},
                                    {'radius', 'diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cylinder is centered along the z-as.
        """
        if self._center is None:
            self._center = False

        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def radius(self) -> float:
        """
        Returns the radius of the cylinder.
        """
        if self._radius is None:
            self._radius = 0.5 * self._diameter

        return self._radius

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def diameter(self) -> float:
        """
        Returns the diameter of the cylinder.
        """
        if self._diameter is None:
            self._diameter = 2.0 * self._radius

        return self._diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height/length of the cylinder.
        """
        if self._height is None:
            self._height = Vector3.distance(self._start_point, self._end_point)

        return self._height

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def start_point(self) -> Vector3:
        """
        Returns the start point of the cylinder.
        """
        if self._start_point is None:
            self._start_point = Vector3(0.0, 0.0, -self._height / 2.0 if self.center else 0.0)

        return self._start_point

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def end_point(self) -> Vector3:
        """
        Returns the end point of the cylinder.
        """
        if self._end_point is None:
            self._end_point = Vector3(0.0, 0.0, self._height / 2.0 if self.center else self._height)

        return self._end_point

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_top(self) -> bool:
        """
        Returns whether the top of the cylinder is extended by eps.
        """
        return self._extend_by_eps_top

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_bottom(self) -> bool:
        """
        Returns whether the bottom of the cylinder is extended by eps.
        """
        return self._extend_by_eps_bottom

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_by_eps_radius(self) -> bool:
        """
        Returns whether the radius of the cylinder is extended by eps.
        """
        return self._extend_by_eps_radius

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._fs

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._fn

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn4n(self) -> bool | None:
        """
        Returns whether to create a circle with multiple of 4 vertices.
        """
        return self._fn4n

    # ------------------------------------------------------------------------------------------------------------------
    def real_fn(self, context: Context) -> int | None:
        """
        Returns the real fixed number of fragments in 360 degrees.
        """
        if self.fn4n:
            return Radius2Sides4n.r2sides4n(context, self.radius)

        return self.fn

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        diameter = self.diameter
        if self.extend_by_eps_radius:
            diameter += 2.0 * context.eps

        height = self.height
        if self.extend_by_eps_top:
            height += context.eps
        if self.extend_by_eps_bottom:
            height += context.eps

        center = self.center and self.extend_by_eps_top == self.extend_by_eps_bottom

        cylinder = PrivateCylinder(height=height,
                                   diameter=diameter,
                                   center=center,
                                   fa=self.fa,
                                   fs=self.fs,
                                   fn=self.real_fn(context))

        if self._explicit_height:
            if not center:
                z = 0.0
                if self.extend_by_eps_bottom:
                    z -= context.eps
                if self.center:
                    z -= 0.5 * self.height
                if z != 0.0:
                    cylinder = Translate3D(z=z, child=cylinder)
        else:
            if self.extend_by_eps_bottom:
                cylinder = Translate3D(z=-context.eps, child=cylinder)

            diff = self.end_point - self.start_point
            cylinder = Translate3D(vector=self.start_point,
                                   child=Rotate3D(angle_y=math.degrees(math.acos(diff.z / diff.length)),
                                                  angle_z=math.degrees(math.atan2(diff.y, diff.x)),
                                                  child=cylinder))

        return cylinder

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Set

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class RotateExtrude(PrivateSingleChildOpenScadCommand):
    """
    Rotational extrusion spins a 2D shape around the Z-axis to form a solid which has rotational symmetry. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#rotate_extrude.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float = 360.0,
                 convexity: int | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param angle: Specifies the number of degrees to sweep, starting at the positive X axis. The direction of the
                      sweep follows the Right-Hand Rule, hence a negative angle sweeps clockwise.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line
                          through the child widget.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        :param child: The 2D child widget.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='rotate_extrude', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'fa': '$fa', 'fs': '$fs', 'fn': '$fn'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return {'angle'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'$fs'}

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle(self) -> float:
        """
        Returns the number of degrees to sweep, starting at the positive X axis. The direction of the sweep follows the
        Right-Hand Rule, hence a negative angle sweeps clockwise.
        """
        return self._args['angle']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the convexity.
        """
        return self._args.get('convexity')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._args.get('fa')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._args.get('fs')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._args.get('fn')

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget


class Modify(ScadSingleChildParent):
    """
    Class for applying modifier characters. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Modifier_Characters#Disable_Modifier
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 disable: bool = False,
                 show_only: bool = False,
                 highlight: bool = False,
                 transparent: bool = False,
                 child: ScadWidget):
        """
        Object constructor.

        :param disable: Whether the child widget is ignored.
        :param show_only: Whether to ignore the rest of the design and use this child widget as design root.
        :param highlight: Whether the child widget is used as usual in the rendering process but also draw it
                          unmodified in transparent pink.
        :param transparent: Whether this child widget is used as usual in the rendering process but draw it in
                            transparent gray (all transformations are still applied to the nodes in this tree).
        :param child: The child widget.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._disable: bool = disable
        """
        Whether the child widget is ignored.
        """

        self._show_only: bool = show_only
        """
        Whether to ignore the rest of the design and use this child widget as design root.
        """

        self._highlight: bool = highlight
        """
        Whether the child widget is used as usual in the rendering process but also draw it unmodified in transparent
        pink.
        """

        self._transparent: bool = transparent
        """
        Whether this child widget is used as usual in the rendering process but draw it in transparent gray (all 
        transformations are still applied to the nodes in this tree).
        """

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def disable(self) -> bool:
        """
        Returns whether this SuperSCAD widget is ignored.
        """
        return self._disable

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def show_only(self) -> bool:
        """
        Returns whether to ignore the rest of the design and use this SuperSCAD widget as design root.
        """
        return self._show_only

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def highlight(self) -> bool:
        """
        Returns whether this SuperSCAD widget is used as usual in the rendering process but also draw it unmodified in
        transparent pink.
        """
        return self._highlight

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def transparent(self) -> bool:
        """
        Returns whether this SuperSCAD widget is used as usual in the rendering process but draw it in transparent gray
        (all transformations are still applied to the nodes in this tree).
        """
        return self._transparent

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        modifiers = ('*' if self.disable else '') + \
                    ('!' if self.show_only else '') + \
                    ('#' if self.highlight else '') + \
                    ('%' if self.transparent else '')

        if modifiers == '':
            return self.child

        return PrivateSingleChildOpenScadCommand(command=modifiers + 'union', args={}, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from pathlib import Path
from typing import Any, Dict

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget


class Surface(PrivateOpenScadCommand):
    """
    Surface reads Heightmap information from text or image files. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#surface.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 path: str | Path,
                 center: bool = False,
                 invert: bool | None = None,
                 convexity: int | None = None):
        """
        Object constructor.

        :param path: The path to the file containing the heightmap data.
        :param center: Whether the object is centered in X- and Y-axis. Otherwise, the object is placed in the positive
                       quadrant.
        :param invert: Whether to invert how the color values of imported images are translated into height values. This
                       has no effect when importing text data files.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        """
        if path is not None:
            path = str(path)

        PrivateOpenScadCommand.__init__(self, command='surface', args=locals())

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_required({'path'}, {'center'})

        # We like to validate here whether the path goes to an exiting readable file, but we need the build context for
        # that. So, we test the existence of the file in the build method.

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'path': 'file'}

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the
        child widget.
        """
        return self._args.get('convexity')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def invert(self) -> bool:
        """
        Returns whether to invert how the color values of imported images are translated into height values. This
        has no effect when importing text data files.
        """
        return self._args['invert']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the object is centered in X- and Y-axis. Otherwise, the object is placed in the positive
        quadrant.
        """
        return self._args['center']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def path(self) -> Path:
        """
        Returns The absolute path or the relative path from the target script to the file that will be imported.
        """
        return Path(self._args['path'])

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        path = self.path
        if not path.is_absolute():
            path = context.target_path.parent.joinpath(path)

        if not path.is_file():
            raise FileNotFoundError(f'File {path} does not exist.')

        return self

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateTranslate import PrivateTranslate
from super_scad.type.Vector3 import Vector3


class Translate3D(ScadSingleChildParent):
    """
    Translates (moves) its child widget along the specified vector. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#translate.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 vector: Vector3 | None = None,
                 x: float | None = None,
                 y: float | None = None,
                 z: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param vector: The vector over which the child widget is translated.
        :param x: The distance over which the child widget is translated along the x-axis.
        :param y: The distance over which the child widget is translated along the y-axis.
        :param z: The distance over which the child widget is translated along the z-axis.
        :param child: The child widget to be translated.
        """
        ScadSingleChildParent.__init__(self, child=child)

        self._vector: Vector3 | None = vector
        """
        The vector over which the child widget is translated.
        """

        self._x: float | None = x
        """
        The distance over which the child widget is translated along the x-axis.
        """

        self._y: float | None = y
        """
        The distance over which the child widget is translated along the y-axis.
        """

        self._z: float | None = z
        """
        The distance over which the child widget is translated along the z-axis.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'vector'}, {'x', 'y', 'z'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector3:
        """
        Returns the vector over which the child widget is translated.
        """
        if self._vector is None:
            self._vector = Vector3(self.x, self.y, self.z)

        return self._vector

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def x(self) -> float:
        """
        Returns distance over which the child widget is translated to along the x-axis.
        """
        if self._x is None:
            if self._vector is not None:
                self._x = self._vector.x
            else:
                self._x = 0.0

        return self._x

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def y(self) -> float:
        """
        Returns distance over which the child widget is translated to along the y-axis.
        """
        if self._y is None:
            if self._vector is not None:
                self._y = self._vector.y
            else:
                self._y = 0.0

        return self._y

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def z(self) -> float:
        """
        Returns distance over which the child widget is translated to along the z-axis.
        """
        if self._z is None:
            if self._vector is not None:
                self._z = self._vector.z
            else:
                self._z = 0.0

        return self._z

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateTranslate(vector=self.vector, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



import re
from typing import Any, List, Tuple

from super_scad.scad import Length
from super_scad.scad.Context import Context
from super_scad.scad.Unit import Unit
from super_scad.type import Vector2, Vector3
from super_scad.type.Color import Color


class Formatter:
    """
    A utility class for formatting and rounding arguments of OpenSCAD commands and variables.
    """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def format(context: Context,
               argument: Any,
               is_angle: bool = False,
               is_length: bool = False,
               is_scale: bool = False,
               unit: Unit | None = None) -> str:
        """
        Returns an argument of the OpenSCAD command.

        :param context: The build context.
        :param argument: The argument of OpenSCAD command.
        :param is_scale: Whether the argument or variable is a scale.
        :param is_length: Whether the argument or variable is a length.
        :param is_angle: Whether the argument or variable is an angle.
        :param unit: The unit of length. Mandatory when the argument or variable is a length.
        """
        if is_length and (isinstance(argument, float) or isinstance(argument, int)):
            argument = Length.convert(float(argument), unit, context.get_unit_length_final())
            argument = str(round(argument, context.length_digits))
            if argument == '-0.0':
                argument = '0.0'
            return argument

        if is_angle and (isinstance(argument, float) or isinstance(argument, int)):
            argument = str(round(float(argument), context.angle_digits))
            if argument == '-0.0':
                argument = '0.0'
            return argument

        if is_scale and (isinstance(argument, float) or isinstance(argument, int)):
            argument = str(round(float(argument), context.scale_digits))
            if argument == '-0.0':
                argument = '0.0'
            return argument

        if isinstance(argument, Vector2):
            return "[{}, {}]".format(Formatter.format(context, argument.x, is_angle, is_length, is_scale, unit),
                                     Formatter.format(context, argument.y, is_angle, is_length, is_scale, unit))

        if isinstance(argument, Vector3):
            return "[{}, {}, {}]".format(
                    Formatter.format(context, argument.x, is_angle, is_length, is_scale, unit),
                    Formatter.format(context, argument.y, is_angle, is_length, is_scale, unit),
                    Formatter.format(context, argument.z, is_angle, is_length, is_scale, unit))

        if isinstance(argument, bool):
            return str(argument).lower()

        if isinstance(argument, str):
            return '"{}"'.format(re.sub(r'([\\\"])', r'\\\1', argument))

        if isinstance(argument, int):
            return str(argument)

        if isinstance(argument, float):
            return str(argument)

        if isinstance(argument, List) or isinstance(argument, Tuple):
            parts = [Formatter.format(context, element, is_angle, is_length, is_scale, unit) for element in argument]

            return '[{}]'.format(', '.join(parts))

        if isinstance(argument, Color):
            return str(argument)

        raise ValueError(f'Can not format argument or variable of type {argument.__class__}.')

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Color import Color


class PrivatePaint(PrivateSingleChildOpenScadCommand):
    """
    Paints a child widget using a specified color and opacity. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#color.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 color: Color,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param color: The color and opacity of the child widget.
        :param child: The child widget to be painted.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='color', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'color': 'c'}

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.d3.private.PrivateCylinder import PrivateCylinder
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.util.Radius2Sides4n import Radius2Sides4n


class Cone(ScadWidget):
    """
    Widget for creating cones. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#cylinder.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float,
                 bottom_radius: float | None = None,
                 bottom_diameter: float | None = None,
                 top_radius: float | None = None,
                 top_diameter: float | None = None,
                 center: bool = False,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 fn4n: bool | None = None):
        """
        Object constructor.

        :param height: The height of the cone.
        :param bottom_radius: The radius at the bottom of the cone.
        :param bottom_diameter: The diameter at the bottom of the cone.
        :param top_radius: The radius at the top of the cone.
        :param top_diameter: The diameter at the top of the cone.
        :param center: Whether the cone is centered in the z-direction.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        :param fn4n: Whether to create a cone with a multiple of 4 vertices.
        """
        ScadWidget.__init__(self)

        self._height: float = height
        """
        The height of the cone.
        """

        self._bottom_radius: float | None = bottom_radius
        """
        The bottom radius of the cone.
        """

        self._bottom_diameter: float | None = bottom_diameter
        """
        The bottom diameter of the cone.
        """

        self._top_radius: float | None = top_radius
        """
        The top radius of the cone.
        """

        self._top_diameter: float | None = top_diameter
        """
        The top diameter of the cone.
        """

        self._center: bool = center
        """
        Whether the cone is centered in the z-direction.
        """

        self._fa: float | None = fa
        """
        The minimum angle (in degrees) of each fragment.
        """

        self._fs: float | None = fs
        """
        The minimum circumferential length of each fragment.
        """

        self._fn: int | None = fn
        """
        The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        """

        self._fn4n: bool | None = fn4n
        """
        Whether to create a cone with a multiple of 4 vertices.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'bottom_radius'}, {'bottom_diameter'})
        validator.validate_exclusive({'top_radius'}, {'top_diameter'})
        validator.validate_exclusive({'fn4n'}, {'fa', 'fs', 'fn'})
        validator.validate_required({'height'},
                                    {'bottom_radius', 'bottom_diameter'},
                                    {'top_radius', 'top_diameter'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cone is centered along the z-as.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def bottom_radius(self) -> float:
        """
        Returns the bottom radius of the cone.
        """
        if self._bottom_radius is None:
            self._bottom_radius = 0.5 * self._bottom_diameter

        return self._bottom_radius

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def bottom_diameter(self) -> float:
        """
        Returns the bottom diameter of the cone.
        """
        if self._bottom_diameter is None:
            self._bottom_diameter = 2.0 * self._bottom_radius

        return self._bottom_diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def top_radius(self) -> float:
        """
        Returns the top radius of the cone.
        """
        if self._top_radius is None:
            self._top_radius = 0.5 * self._top_diameter

        return self._top_radius

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def top_diameter(self) -> float:
        """
        Returns the top diameter of the cone.
        """
        if self._top_diameter is None:
            self._top_diameter = 2.0 * self._top_radius

        return self._top_diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height of the cone.
        """
        return self._height

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self._fs

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._fn

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn4n(self) -> bool | None:
        """
        Returns whether to create a circle with multiple of 4 vertices.
        """
        return self._fn4n

    # ------------------------------------------------------------------------------------------------------------------
    def real_fn(self, context: Context) -> int | None:
        """
        Returns the real fixed number of fragments in 360 degrees.
        """
        if self.fn4n:
            return Radius2Sides4n.r2sides4n(context, max(self.bottom_radius, self.top_radius))

        return self.fn

    # ----------------------------------------f--------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateCylinder(height=self.height,
                               bottom_diameter=self.bottom_diameter,
                               top_diameter=self.top_diameter,
                               center=self.center,
                               fa=self.fa,
                               fs=self.fs,
                               fn=self.real_fn(context))

# ----------------------------------------------------------------------------------------------------------------------
